#include <iostream>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>



int main()
{
	std::cout << "TEST: recv() on a listen socket" << std::endl;
	auto socket = ::socket(AF_INET, SOCK_STREAM, 0);
	if (socket == -1) {
		std::cout << "Unexpected error creating the socket" << std::endl << errno << ": "
			<< strerror(errno) << std::endl;
		return 1;
	}

	// Bind to address
	struct sockaddr_in server_address;
	server_address.sin_family = AF_INET;
	server_address.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	int bind_result = ::bind(socket, reinterpret_cast<struct sockaddr*>(&server_address),
		sizeof(server_address));
	if (bind_result == -1) {
		std::cout << "Unexpected error binding the socket" << std::endl << errno << ": "
				  << strerror(errno) << std::endl;
		return 2;
	}

	// Listen is apparently required before getsockname will work.
	if (::listen(socket, 32) == -1) {
		std::cout << "Unexpected error setting the socket to listen" << std::endl << errno << ": "
			<< strerror(errno) << std::endl;
		return 3;
	}

	struct sockaddr_in peeraddr;
	socklen_t peeraddrlen;
	auto peer_result = getpeername(socket, (struct sockaddr *) &peeraddr, &peeraddrlen);
	if (peer_result == -1 && errno == ENOTCONN) {
		// Is this expected?
		std::cout << "Socket is not connected, for the purposes of getpeername" << std::endl
			<< errno << ": " << strerror(errno) << std::endl;
	} else if (peer_result == -1){
		std::cout << "Unexpected error in getpeername" << std::endl << errno << ": "
			<< strerror(errno) << std::endl;
		return 4;
	} else {
		std::cout << "getpeername unexpectedly succeeded!" << std::endl;
		return 5;
	}

	auto status = fcntl(socket, F_SETFL, fcntl(socket, F_GETFL, 0) | O_NONBLOCK);
	if (status == -1){
		std::cout << "Unexpected error setting socket to non-blocking" << std::endl << errno
			<< ": " << strerror(errno) << std::endl;
		return 6;
	}

	char buffer[1];
	auto recv_result = ::recv(socket, static_cast<void*>(&buffer), 1, 0);
	if (recv_result == -1 && errno == ENOTCONN) {
		std::cout << "recv() fails because the socket is not connected. Not sure if this is "
					 "expected, but at least it is consistent" << std::endl;
	} else if (recv_result == -1 && errno == EAGAIN) {
		std::cout << "recv() fails because it would block. This is inconsistent with with the "
					 "outcome of getpeername() above." << std::endl;
	} else if (recv_result == -1) {
		std::cout << "Unexpected error with recv()" << std::endl << errno
				  << ": " << strerror(errno) << std::endl;
		return 7;
	} else {
		std::cout << "Unexpected success for recv()" << std::endl;
		return 8;
	}

    return 0;
}
